package pp.s1184725.boppi;

import java.util.*;

public class CachingSymbolTable<T> {
	protected Stack<Map<String,Variable<T>>> symbolMapStack;
	protected Stack<Integer> offsets;
	protected Map<String,Variable<T>> symbolCache;
	protected int offset;

	/**
	 * Creates an empty symbol table with no open scope.
	 */
	public CachingSymbolTable() {
		symbolMapStack = new Stack<>();
		symbolCache = new HashMap<>();
		offsets = new Stack<>();
		offset = 0;
	}
	
	/**
	 * Opens a lexical scope.
	 */
	public void openScope() {
		symbolMapStack.push(new HashMap<>());
		offsets.push(offset);
	}
	
	/**
	 * Closes a lexical scope, removing all definitions within this scope.
	 * @throws EmptyStackException if no scope is open
	 */
	public void closeScope() throws EmptyStackException {
		Map<String,Variable<T>> deletions = symbolMapStack.pop();
		for (String identifier : deletions.keySet()) {
			Optional<Variable<T>> maybeType = symbolMapStack.stream().filter((map) -> map.containsKey(identifier)).map((map) -> map.get(identifier)).reduce((__, snd) -> snd);
			maybeType.ifPresent((var) -> symbolCache.replace(identifier, var));
			if (!maybeType.isPresent())
				symbolCache.remove(identifier);
		}
		offset = offsets.pop();
	}
	
	/**
	 * Associates an identifier with a certain type in the current lexical scope.
	 * @param id the name of the identifier
	 * @param type the type of the identifier
	 * @return the type of the identifier
	 * @throws Exception if the identifier is declared in the current scope already
	 * @throws EmptyStackException if no scope is open
	 */
	public Variable<T> put(String id, T type) throws Exception, EmptyStackException {
		if (symbolMapStack.peek().containsKey(id))
			throw new Exception(String.format("Identifier '%s' already declared in this scope.", id));
		
		Variable<T> var = new Variable<>(type, offset);
		// TODO refactor to get type size
		offset += ((SimpleType)type).getSize();
		symbolMapStack.peek().put(id, var);
		symbolCache.put(id, var);
		return var;
	}
	
	/**
	 * Returns whether the given identifier is defined.
	 * @param id the name of the identifier
	 * @return true if the identifier has a defined type
	 * @throws EmptyStackException if no scope is open
	 */
	public boolean has(String id) throws EmptyStackException {
		if (symbolMapStack.isEmpty())
			throw new EmptyStackException();

		return symbolCache.containsKey(id);
	}
	
	/**
	 * Retrieves the type of an identifier, if any.
	 * @param id the name of the identifier
	 * @return the type of the identifier
	 * @throws Exception if the identifier is not defined
	 * @throws EmptyStackException if no scope is open
	 */
	public Variable<T> get(String id) throws Exception, EmptyStackException {
		if (!this.has(id))
			throw new Exception(String.format("Identifier '%s' is undefined.", id));

		return symbolCache.get(id);
	}
}
